/* aln320.c: Fast Ethernet driver for Linux. */
/*
	
	This software may be used and distributed according to the terms
	of the GNU Public License, incorporated herein by reference.
    All other rights reserved.

	This driver is for boards based on the p3043 and aln320 PCI ethernet
	chips.

		
*/

static const char *version =
"aln320.c:v1.10 Mar/05/99 \n";

/* A few user-configurable values. */

/* Size of the in-memory receive ring. */
#define RX_BUF_LEN_IDX	3			/* 0==8K, 1==16K, 2==32K, 3==64K */
#define RX_BUF_LEN (8192 << RX_BUF_LEN_IDX)

/* Size of the Tx bounce buffers -- must be at least (dev->mtu+14+4). */
#define TX_BUF_SIZE	1536
#define RX_BUF_SIZE	1536

/* PCI Tuning Parameters
   Threshold is bytes transferred to chip before transmission starts. */
#define TX_FIFO_THRESH 256	/* In bytes, rounded down to 32 byte units. */

/* The following settings are log_2(bytes)-4:  0 == 16 bytes .. 6==1024. */
#define RX_FIFO_THRESH	4		/* Rx buffer level before first PCI xfer.  */
#define RX_DMA_BURST	4		/* Maximum PCI burst, '4' is 256 bytes */
#define TX_DMA_BURST	4

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  ((2000*HZ)/1000)

#include <linux/config.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/bios32.h>
#include <asm/processor.h>		/* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

/* define all ioaddr */

#define byPAR0                       ioaddr
#define byRCR                        ioaddr + 6
#define byTCR                        ioaddr + 7
#define byCR0                        ioaddr + 8
#define byCR1                        ioaddr + 9
#define byISR0                       ioaddr + 0x0c
#define byISR1                       ioaddr + 0x0d
#define byIMR0                       ioaddr + 0x0e
#define byIMR1                       ioaddr + 0x0f
#define byMAR0                       ioaddr + 0x10
#define byMAR1                       ioaddr + 0x11
#define byMAR2                       ioaddr + 0x12
#define byMAR3                       ioaddr + 0x13
#define byMAR4                       ioaddr + 0x14
#define byMAR5                       ioaddr + 0x15
#define byMAR6                       ioaddr + 0x16
#define byMAR7                       ioaddr + 0x17
#define dwCurrentRxDescAddr          ioaddr + 0x18
#define dwCurrentTxDescAddr          ioaddr + 0x1c
#define dwCurrentRDSE0               ioaddr + 0x20
#define dwCurrentRDSE1               ioaddr + 0x24
#define dwCurrentRDSE2               ioaddr + 0x28
#define dwCurrentRDSE3               ioaddr + 0x2c
#define dwNextRDSE0                  ioaddr + 0x30
#define dwNextRDSE1                  ioaddr + 0x34
#define dwNextRDSE2                  ioaddr + 0x38
#define dwNextRDSE3                  ioaddr + 0x3c
#define dwCurrentTDSE0               ioaddr + 0x40
#define dwCurrentTDSE1               ioaddr + 0x44
#define dwCurrentTDSE2               ioaddr + 0x48
#define dwCurrentTDSE3               ioaddr + 0x4c
#define dwNextTDSE0                  ioaddr + 0x50
#define dwNextTDSE1                  ioaddr + 0x54
#define dwNextTDSE2                  ioaddr + 0x58
#define dwNextTDSE3                  ioaddr + 0x5c
#define dwCurrRxDMAPtr               ioaddr + 0x60
#define dwCurrTxDMAPtr               ioaddr + 0x64
#define byMPHY                       ioaddr + 0x6c
#define byMIISR                      ioaddr + 0x6d
#define byBCR0                       ioaddr + 0x6e
#define byBCR1                       ioaddr + 0x6f
#define byMIICR                      ioaddr + 0x70
#define byMIIAD                      ioaddr + 0x71
#define wMIIDATA                     ioaddr + 0x72
#define byEECSR                      ioaddr + 0x74
#define byTEST                       ioaddr + 0x75
#define byGPIO                       ioaddr + 0x76
#define byCFGA                       ioaddr + 0x78
#define byCFGB                       ioaddr + 0x79
#define byCFGC                       ioaddr + 0x7a
#define byCFGD                       ioaddr + 0x7b
#define wTallyCntMPA                 ioaddr + 0x7c
#define wTallyCntCRC                 ioaddr + 0x7d
/*---------------------  Exioaddr Definitions -------------------------*/

//

// Bits in the RCR register

//

#define RCR_RRFT2           0x80        //
#define RCR_RRFT1           0x40        //
#define RCR_RRFT0           0x20        //
#define RCR_PROM            0x10        //
#define RCR_AB              0x08        //
#define RCR_AM              0x04        //
#define RCR_AR              0x02        //
#define RCR_SEP             0x01        //
//

// Bits in the TCR register

//

#define TCR_RTSF            0x80        //
#define TCR_RTFT1           0x40        //
#define TCR_RTFT0           0x20        //
#define TCR_OFSET           0x08        //
#define TCR_LB1             0x04        // loopback[1]
#define TCR_LB0             0x02        // loopback[0]
//

// Bits in the CR0 register

//

#define CR0_RDMD            0x40        // rx descriptor polling demand
#define CR0_TDMD            0x20        // tx descriptor polling demand
#define CR0_TXON            0x10        //
#define CR0_RXON            0x08        //
#define CR0_STOP            0x04        // stop NIC, default = 1
#define CR0_STRT            0x02        // start NIC
#define CR0_INIT            0x01        // start init process


//

// Bits in the CR1 register

//

#define CR1_SFRST           0x80        // software reset
#define CR1_RDMD1           0x40        // RDMD1     
#define CR1_TDMD1           0x20        // TDMD1   
#define CR1_KEYPAG          0x10        // turn on par/key
#define CR1_DPOLL           0x08        // disable rx/tx auto polling
#define CR1_FDX             0x04        // full duplex mode
#define CR1_ETEN            0x02        // early tx mode
#define CR1_EREN            0x01        // early rx mode

// Bits in the CR register

//

#define CR_RDMD            0x0040        // rx descriptor polling demand
#define CR_TDMD            0x0020        // tx descriptor polling demand
#define CR_TXON            0x0010        //
#define CR_RXON            0x0008        //
#define CR_STOP            0x0004        // stop NIC, default = 1
#define CR_STRT            0x0002        // start NIC
#define CR_INIT            0x0001        // start init process
#define CR_SFRST           0x8000        // software reset
#define CR_RDMD1           0x4000        // RDMD1     
#define CR_TDMD1           0x2000        // TDMD1   
#define CR_KEYPAG          0x1000        // turn on par/key
#define CR_DPOLL           0x0800        // disable rx/tx auto polling
#define CR_FDX             0x0400        // full duplex mode
#define CR_ETEN            0x0200        // early tx mode
#define CR_EREN            0x0100      // early rx mode


//

// Bits in the IMR0 register

//

#define IMR0_CNTM           0x80        //
#define IMR0_BEM            0x40        //
#define IMR0_RUM            0x20        //
#define IMR0_TUM            0x10        //
#define IMR0_TXEM           0x08        //
#define IMR0_RXEM           0x04        //
#define IMR0_PTXM           0x02        //
#define IMR0_PRXM           0x01        //



// define imrshadow

#define IMRShadow         0x5AFF

// Bits in the IMR1 register

//

#define IMR1_INITM          0x80        //
#define IMR1_SRCM           0x40        //
#define IMR1_NBFM           0x10        //
#define IMR1_PRAIM          0x08        //
#define IMR1_RES0M          0x04        //
#define IMR1_ETM            0x02        //
#define IMR1_ERM            0x01        //



//

// Bits in the ISR register

//

#define ISR_INITI           0x8000      //
#define ISR_SRCI            0x4000      //
#define ISR_ABTI            0x2000      //
#define ISR_NORBF           0x1000      //
#define ISR_PKTRA           0x0800      //
#define ISR_RES0            0x0400      //
#define ISR_ETI             0x0200      //
#define ISR_ERI             0x0100      //
#define ISR_CNT             0x0080      //
#define ISR_BE              0x0040      //
#define ISR_RU              0x0020      //
#define ISR_TU              0x0010      //
#define ISR_TXE             0x0008      //
#define ISR_RXE             0x0004      //
#define ISR_PTX             0x0002      //
#define ISR_PRX             0x0001      //



//

// Bits in the ISR0 register

//

#define ISR0_CNT            0x80        //
#define ISR0_BE             0x40        //
#define ISR0_RU             0x20        //
#define ISR0_TU             0x10        //
#define ISR0_TXE            0x08        //
#define ISR0_RXE            0x04        //
#define ISR0_PTX            0x02        //
#define ISR0_PRX            0x01        //



//

// Bits in the ISR1 register

//

#define ISR1_INITI          0x80        //
#define ISR1_SRCI           0x40        //
#define ISR1_NORBF          0x10        //
#define ISR1_PKTRA          0x08        //
#define ISR1_ETI            0x02        //
#define ISR1_ERI            0x01        //

// ISR ABNORMAL CONDITION

#define ISR_ABNORMAL ISR_BE+ISR_RU+ISR_TU+ISR_CNT+ISR_NORBF+ISR_PKTRA      



//

// Bits in the MIISR register

//

#define MIISR_MIIERR        0x08        //
#define MIISR_MRERR         0x04        //
#define MIISR_LNKFL         0x02        //
#define MIISR_SPEED         0x01        //


//

// Bits in the MIICR register

//

#define MIICR_MAUTO         0x80        //
#define MIICR_RCMD          0x40        //
#define MIICR_WCMD          0x20        //
#define MIICR_MDPM          0x10        //
#define MIICR_MOUT          0x08        //
#define MIICR_MDO           0x04        //
#define MIICR_MDI           0x02        //
#define MIICR_MDC           0x01        //

//

// Bits in the EECSR register

//

#define EECSR_EEPR          0x80        // eeprom programed status, 73h means programed
#define EECSR_EMBP          0x40        // eeprom embeded programming
#define EECSR_AUTOLD        0x20        // eeprom content reload
#define EECSR_DPM           0x10        // eeprom direct programming
#define EECSR_CS            0x08        // eeprom CS pin
#define EECSR_SK            0x04        // eeprom SK pin
#define EECSR_DI            0x02        // eeprom DI pin
#define EECSR_DO            0x01        // eeprom DO pin



//

// Bits in the BCR0 register

//

#define BCR0_CRFT2          0x20        //
#define BCR0_CRFT1          0x10        //
#define BCR0_CRFT0          0x08        //
#define BCR0_DMAL2          0x04        //
#define BCR0_DMAL1          0x02        //
#define BCR0_DMAL0          0x01        //

//

// Bits in the BCR1 register

//

#define BCR1_CTSF           0x20        //
#define BCR1_CTFT1          0x10        //
#define BCR1_CTFT0          0x08        //
#define BCR1_POT2           0x04        //
#define BCR1_POT1           0x02        //
#define BCR1_POT0           0x01        //



//

// Bits in the CFGA register

//

#define CFGA_EELOAD         0x80        // enable eeprom embeded and direct programming
#define CFGA_JUMPER         0x40        //
#define CFGA_MTGPIO         0x08        //
#define CFGA_T10EN          0x02        //
#define CFGA_AUTO           0x01        //



//

// Bits in the CFGB register

//

#define CFGB_PD             0x80        //
#define CFGB_POLEN          0x02        //
#define CFGB_LNKEN          0x01        //



//

// Bits in the CFGC register

//

#define CFGC_M10TIO         0x80        //
#define CFGC_M10POL         0x40        //
#define CFGC_PHY1           0x20        //
#define CFGC_PHY0           0x10        //
#define CFGC_BTSEL          0x08        //
#define CFGC_BPS2           0x04        // bootrom select[2]
#define CFGC_BPS1           0x02        // bootrom select[1]
#define CFGC_BPS0           0x01        // bootrom select[0]



//

// Bits in the CFGD register

//

#define CFGD_GPIOEN         0x80        //
#define CFGD_DIAG           0x40        //
#define CFGD_MAGIC          0x10        //
#define CFGD_CFDX           0x04        //
#define CFGD_CEREN          0x02        //
#define CFGD_CETEN          0x01        //

// Bits in RSR
#define RSR_RERR            0x00000001 
#define RSR_CRC             0x00000002
#define RSR_FAE             0x00000004
#define RSR_FOV             0x00000008 
#define RSR_LONG            0x00000010
#define RSR_RUNT            0x00000020
#define RSR_SERR            0x00000040
#define RSR_BUFF            0x00000080
#define RSR_EDP             0x00000100
#define RSR_STP             0x00000200
#define RSR_CHN             0x00000400
#define RSR_PHY             0x00000800
#define RSR_BAR             0x00001000
#define RSR_MAR             0x00002000
#define RSR_RXOK            0x00008000         
#define RSR_ABNORMAL        RSR_RERR+RSR_LONG+RSR_RUNT 
 //Bits in TSR

#define TSR_NCR0            0x00000001
#define TSR_NCR1            0x00000002
#define TSR_NCR2            0x00000004
#define TSR_NCR3            0x00000008
#define TSR_COLS            0x00000010
#define TSR_CDH             0x00000080
#define TSR_ABT             0x00000100
#define TSR_OWC             0x00000200
#define TSR_CRS             0x00000400
#define TSR_UDF             0x00000800
#define TSR_TBUFF           0x00001000
#define TSR_SERR            0x00002000
#define TSR_JAB             0x00004000
#define TSR_TERR            0x00008000
#define TSR_ABNORMAL        TSR_TERR+TSR_OWC+TSR_ABT+TSR_JAB+TSR_CRS 
#define TSR_OWN_BIT         0x80000000 
  
  
#define CB_DELAY_LOOP_WAIT  (10)        // 10ms
// enabled mask value of irq

#define W_IMR_MASK_VALUE    0x1BFF      // initial value of IMR

// Ethernet address filter type
#define PKT_TYPE_DIRECTED           0x0001  // obselete, directed address is always accepted
#define PKT_TYPE_MULTICAST          0x0002
#define PKT_TYPE_ALL_MULTICAST      0x0004
#define PKT_TYPE_BROADCAST          0x0008
#define PKT_TYPE_PROMISCUOUS        0x0020
//#define PKT_TYPE_LONG               0x2000
#define PKT_TYPE_RUNT               0x4000
#define PKT_TYPE_ERROR              0x8000  // accept error packets, e.g. CRC error



// Loopback mode

#define NIC_LB_NONE         0x00        //
#define NIC_LB_INTERNAL     0x01        //
#define NIC_LB_PHY          0x02        // MII or Internal-10BaseT loopback



#define TX_RING_SIZE    8 	
#define RX_RING_SIZE    20	
#define PKT_BUF_SZ		1536			/* Size of each temporary Rx buffer.*/



    /* Define macros for converting virtual to physical addresses */

 /* Transmit and receive descriptors defenition */

struct aln320_tx_desc 
    {
    union VTC_tx_status_tag
        {
        struct
            {
            ulong ncro                      : 1;
            ulong ncr1                      : 1;
            ulong ncr2                      : 1;
            ulong ncr3                      : 1;
            ulong cols                      : 1;
            ulong reserve_1                 : 2;
            ulong cdh                       : 1;
            ulong abt                       : 1;
            ulong owc                       : 1;
            ulong crs                       : 1;
            ulong udf                       : 1;
            ulong tbuff                     : 1;
            ulong serr                      : 1;
            ulong jab                       : 1;
            ulong terr                      : 1;
            ulong reserve_2                 : 15; 
            ulong own_bit                   : 1;
            } bits;
        ulong lw;
        } tx_status;

    union VTC_tx_ctrl_tag
        {
        struct
            {
            ulong tx_buf_size               :11;
            ulong extend_tx_buf_size        : 4;
            ulong chn                       : 1;
            ulong crc                       : 1;
            ulong reserve_1                 : 4;
            ulong stp                       : 1;
            ulong edp                       : 1;
            ulong ic                        : 1;
            ulong reserve_2                 : 8; 
            } bits ;
        ulong lw;
        } tx_ctrl;

    ulong buf_addr_1          :32;
    ulong buf_addr_2          :32;

    };



struct aln320_rx_desc
    {
    union VTC_rx_status_tag
        {
        struct
            {
            ulong rerr                      : 1;
            ulong crc_error                 : 1;
            ulong fae                       : 1;
            ulong fov                       : 1;
            ulong toolong                   : 1;
            ulong runt                      : 1;
            ulong serr                      : 1;
            ulong buff                      : 1;
            ulong edp                       : 1;
            ulong stp                       : 1;
            ulong chn                       : 1;
            ulong phy                       : 1;
            ulong bar                       : 1;
            ulong mar                       : 1;
            ulong reserve_1                 : 1;
            ulong rxok                      : 1;
            ulong frame_length              : 11;
            ulong reverve_2                 : 4;
            ulong own_bit                   : 1;
            } bits;
        ulong lw;
        } rx_status;

    union VTC_rx_ctrl_tag
        {
        struct
            {
            ulong rx_buf_size               :11;
            ulong extend_rx_buf_size           :4;
            ulong reserved_1                :17;
            } bits;
        ulong lw;
        } rx_ctrl;

    ulong buf_addr_1                    :32;
    ulong buf_addr_2                    :32;

    };


/* Kernel compatibility defines, common to David Hind's PCMCIA package.
   This is only in the support-all-kernels source code. */
#include <linux/version.h>		/* Evil, but neccessary */

#if defined (LINUX_VERSION_CODE) && LINUX_VERSION_CODE < 0x10300
#define RUN_AT(x) (x)			/* What to put in timer->expires.  */
#define DEV_ALLOC_SKB(len) alloc_skb(len, GFP_ATOMIC)
#define virt_to_bus(addr)  ((unsigned long)addr)
#define bus_to_virt(addr) ((void*)addr)

#else  /* 1.3.0 and later */
#define RUN_AT(x) (jiffies + (x))
#define DEV_ALLOC_SKB(len) dev_alloc_skb(len + 2)
#endif

#if defined (LINUX_VERSION_CODE) && LINUX_VERSION_CODE < 0x10338
#ifdef MODULE
#if !defined(CONFIG_MODVERSIONS) && !defined(__NO_VERSION__)
char kernel_version[] = UTS_RELEASE;
#endif
#else
#undef MOD_INC_USE_COUNT
#define MOD_INC_USE_COUNT
#undef MOD_DEC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#endif /* 1.3.38 */

#if (LINUX_VERSION_CODE >= 0x10344)
#define NEW_MULTICAST
#include <linux/delay.h>
#endif

#ifdef SA_SHIRQ
#define FREE_IRQ(irqnum, dev) free_irq(irqnum, dev)
#define REQUEST_IRQ(i,h,f,n, instance) request_irq(i,h,f,n, instance)
#define IRQ(irq, dev_id, pt_regs) (irq, dev_id, pt_regs)
#else
#define FREE_IRQ(irqnum, dev) free_irq(irqnum)
#define REQUEST_IRQ(i,h,f,n, instance) request_irq(i,h,f,n)
#define IRQ(irq, dev_id, pt_regs) (irq, pt_regs)
#endif

#if (LINUX_VERSION_CODE < 0x20123)
#define test_and_set_bit(val, addr) set_bit(val, addr)
#endif

/* The I/O extent. */
#define p3043_TOTAL_SIZE 0x80

#ifdef HAVE_DEVLIST
struct netdev_entry aln320_drv =
{"aln320", aln320_probe, p3043_TOTAL_SIZE, NULL};
#endif

static int p3043_debug = 1;

/*
				Theory of Operation

I. Board Compatibility

This device driver is designed for the RealTek p3043, the RealTek Fast
Ethernet controllers for PCI.  This chip is used on a few clone boards.


II. Board-specific settings

PCI bus devices are configured by the system at boot time, so no jumpers
need to be set on the board.  The system BIOS will assign the
PCI INTA signal to a (preferably otherwise unused) system IRQ line.
Note: Kernel versions earlier than 1.3.73 do not support shared PCI
interrupt lines.

III. Driver operation

IIIa. Rx Ring buffers

The receive unit uses a single linear ring buffer rather than the more
common (and more efficient) descriptor-based architecture.  Incoming frames
are sequentially stored into the Rx region, and the host copies them into
skbuffs.

Comment: While it is theoretically possible to process many frames in place,
any delay in Rx processing would cause us to drop frames.  More importantly,
the Linux protocol stack is not designed to operate in this manner.

IIIb. Tx operation

The p3043 uses a fixed set of four Tx descriptors in register space.
In a stunningly bad design choice, Tx frames must be 32 bit aligned.  Linux
aligns the IP header on word boundaries, and 14 byte ethernet header means
that almost all frames will need to be copied to an alignment buffer.

IVb. References

http://www.via.com.tw
http://cesdis.gsfc.nasa.gov/linux/misc/NWay.html

IVc. Errata

*/

#define PCI_VENDOR_ID_VIA		0x1106
#define PCI_DEVICE_ID_VIA_3043	        0x3043

/* The rest of these values should never change. */
#define NUM_TX_DESC	4			/* Number of Tx descriptor registers. */



struct p3043_private {
	char devname[8];			/* Used only for kernel debugging. */
	const char *product_name;
        struct aln320_rx_desc *rx_ring;
	struct aln320_tx_desc *tx_ring;
        char *rx_buffs[RX_RING_SIZE];
        char *tx_buffs[TX_RING_SIZE];

	/* temporary Rx buffers. */

	struct device *next_module;
	int chip_id;
	int chip_revision;
	struct enet_statistics stats;
	struct timer_list timer;	/* Media selection timer. */
	unsigned int cur_rx, cur_tx;	/* The next free and used entries */
	unsigned int dirty_rx, dirty_tx;
	/* The saved address of a sent-in-place packet/buffer, for skfree(). */
	struct sk_buff* tx_skbuff[TX_RING_SIZE];
	unsigned char mc_filter[8];			/* Current multicast filter. */
	char phys[4];						/* MII device addresses. */
	unsigned int tx_full:1;				/* The Tx queue is full. */
	unsigned int full_duplex:1;			/* Full-duplex operation requested. */
	unsigned int default_port:4;		/* Last dev->if_port value. */
	unsigned int media2:4;				/* Secondary monitored media port. */
	unsigned int medialock:1;			/* Don't sense media type. */
	unsigned int mediasense:1;			/* Media sensing in progress. */
};

#ifdef MODULE
/* Used to pass the full-duplex flag, etc. */
static int options[] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int full_duplex[] = {-1, -1, -1, -1, -1, -1, -1, -1};
#if LINUX_VERSION_CODE > 0x20118
MODULE_AUTHOR("Donald Becker <becker@cesdis.gsfc.nasa.gov>");
MODULE_DESCRIPTION("ALN-320 Fast Ethernet driver");
MODULE_PARM(debug, "i");
MODULE_PARM(options, "1-" __MODULE_STRING(8) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(8) "i");
MODULE_PARM(max_interrupt_work, "i");
#endif
#endif

static struct device *p3043_probe1(struct device *dev, int ioaddr, int irq,
									 int chip_id, int options, int card_idx);
static int p3043_open(struct device *dev);
static void p3043_timer(unsigned long data);
static void p3043_tx_timeout(struct device *dev);
static void p3043_init_ring(struct device *dev);
static int p3043_start_xmit(struct sk_buff *skb, struct device *dev);
static int p3043_rx(struct device *dev);
static void p3043_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
static int p3043_close(struct device *dev);
static struct enet_statistics *p3043_get_stats(struct device *dev);
#ifdef NEW_MULTICAST
static void set_rx_mode(struct device *dev);
#else
static void set_rx_mode(struct device *dev, int num_addrs, void *addrs);
#endif
static int QueryAuto(int);
static int  ReadMII(int byMIIIndex,int);
static void WriteMII(char ,char ,char ,int);
static void MIIDelay(void);
void post_reset_delay(void);
uint read_tick_counter(void);

#ifdef MODULE
/* A list of all installed p3043 devices, for removing the driver module. */
static struct device *root_p3043_dev = NULL;
#endif
int j;
int aln320_probe(struct device *dev)
{
	int cards_found = 0;
	static int pci_index = 0;	/* Static, for multiple probe calls. */

	/* Ideally we would detect all network cards in slot order.  That would
	   be best done a central PCI probe dispatch, which wouldn't work
	   well with the current structure.  So instead we detect just the
	   Rtl81*9 cards in slot order. */

	if (pcibios_present()) {
		unsigned char pci_bus, pci_device_fn;

		for (;pci_index < 0xff; pci_index++) {
			unsigned char pci_irq_line, pci_latency;
			unsigned short pci_command, vendor, device;
			unsigned int pci_ioaddr;

			if (pcibios_find_class (PCI_CLASS_NETWORK_ETHERNET << 8,
#ifdef REVERSE_PROBE_ORDER
									0xff - pci_index,
#else
									pci_index,
#endif
									&pci_bus, &pci_device_fn)
				!= PCIBIOS_SUCCESSFUL)
				break;
			pcibios_read_config_word(pci_bus, pci_device_fn,
									 PCI_VENDOR_ID, &vendor);
			if (vendor != PCI_VENDOR_ID_VIA)
				continue;


			pcibios_read_config_word(pci_bus, pci_device_fn,
									 PCI_DEVICE_ID, &device);
			pcibios_read_config_byte(pci_bus, pci_device_fn,
									 PCI_INTERRUPT_LINE, &pci_irq_line);
			pcibios_read_config_dword(pci_bus, pci_device_fn,
									  PCI_BASE_ADDRESS_0, &pci_ioaddr);
			/* Remove I/O space marker in bit 0. */
			pci_ioaddr &= ~3;

			if (device != PCI_DEVICE_ID_VIA_3043
				&&  device != PCI_DEVICE_ID_VIA_3043) {
				printk("Unknown RealTek PCI ethernet chip type %4.4x detected:"
					   " not configured.\n", device);
				continue;
			}
			if (check_region(pci_ioaddr, p3043_TOTAL_SIZE))
				continue;

#ifdef MODULE
			dev = p3043_probe1(dev, pci_ioaddr, pci_irq_line, device,
								 options[cards_found], cards_found);
#else
			dev = p3043_probe1(dev, pci_ioaddr, pci_irq_line, device,
								 dev ? dev->mem_start : 0, -1);
#endif

			if (dev) {
				/* Get and check the bus-master and latency values. */
				pcibios_read_config_word(pci_bus, pci_device_fn,
										 PCI_COMMAND, &pci_command);
				if ( ! (pci_command & PCI_COMMAND_MASTER)) {
					printk("  PCI Master Bit has not been set! Setting...\n");
					pci_command |= PCI_COMMAND_MASTER;
					pcibios_write_config_word(pci_bus, pci_device_fn,
											  PCI_COMMAND, pci_command);
				}
				pcibios_read_config_byte(pci_bus, pci_device_fn,
										 PCI_LATENCY_TIMER, &pci_latency);
				if (pci_latency < 10) {
					printk("  PCI latency timer (CFLT) is unreasonably low "
						   "at %d.  Setting to 64 clocks.\n", pci_latency);
					pcibios_write_config_byte(pci_bus, pci_device_fn,
											  PCI_LATENCY_TIMER, 64);
				} else if (p3043_debug > 1)
					printk("  PCI latency timer (CFLT) is %#x.\n",
						   pci_latency);
				dev = 0;
				cards_found++;
			}
		}
	}

#if defined (MODULE)
	return cards_found;
#else
	return cards_found ? 0 : -ENODEV;
#endif
}

static struct device *p3043_probe1(struct device *dev, int ioaddr, int irq,
								   int chip_id, int options, int card_idx)
{
	static int did_version = 0;			/* Already printed version info. */
	struct p3043_private *tp;
	int   i,x;
        int   FDXFlag;
        int   byMIIvalue,LineSpeed,MIICRbak;
	if (p3043_debug > 0  &&  did_version++ == 0)
		printk(version);
	dev = init_etherdev(dev, 0);


	/* Perhaps this should be read from the EEPROM? */
	for (i = 0; i <6; i++)
		dev->dev_addr[i] = inb(byPAR0 + i);
        printk("Ethernet Address: ");
	for (i = 0; i < 5; i++)
		printk("%2.2x:", dev->dev_addr[i]);
	printk("%2.2x\n", dev->dev_addr[i]);


        /* restart MII auto-negotiation*/
             WriteMII(0,9,1,ioaddr);
             printk("Analyzing Media type,this will take several seconds........");
             for(i=0;i<5;i++)   { 
                for(x=0;x<100;x++)  
                  post_reset_delay();  
                if(ReadMII(1,ioaddr)&0x0020)
                  break;
             }  
             printk("OK\n");
 /* query MII to know LineSpeed,duplex mode */
              byMIIvalue =inb(ioaddr+0x6d);
              LineSpeed=byMIIvalue & MIISR_SPEED;
              if (LineSpeed==1){
              printk("Linespeed=10Mbs");
              }else{
              printk("Linespeed=100Mbs");
              }
              FDXFlag=QueryAuto(ioaddr);
              if (FDXFlag==1){
              printk(" Fullduplex\n");
               outw(CR_FDX,byCR0);
              }else{
              printk(" Halfduplex\n");
              }

/* set MII 10 FULL ON */
             WriteMII(17,1,1,ioaddr); 

/* turn on MII link change */
             MIICRbak=inb(byMIICR);
             outb(MIICRbak&0x7F,byMIICR);
	     MIIDelay(); 
             outb(0x41,byMIIAD);
	     MIIDelay(); 
              
          //   while((inb(byMIIAD)&0x20)==0)
          //               ;              
        outb(MIICRbak|0x80,byMIICR);

	dev->base_addr = ioaddr;
	dev->irq = irq;


	/* We do a request_region() to register /proc/ioports info. */
	request_region(ioaddr, p3043_TOTAL_SIZE, "VIA p3043 Fast Ethernet");

	/* Some data structures must be quadword aligned. */
	tp = kmalloc(sizeof(*tp), GFP_KERNEL | GFP_DMA);
	memset(tp, 0, sizeof(*tp));
	dev->priv = tp;

#ifdef MODULE
	tp->next_module = root_p3043_dev;
	root_p3043_dev = dev;
#endif

	tp->chip_id = chip_id;

	tp->phys[0] = -1;

	/* The lower four bits are the media type. */
	if (options > 0) {
		tp->full_duplex = (options & 16) ? 1 : 0;
		tp->default_port = options & 15;
		if (tp->default_port)
			tp->medialock = 1;
	}
#ifdef MODULE
	if (card_idx >= 0) {
		if (full_duplex[card_idx] >= 0)
			tp->full_duplex = full_duplex[card_idx];
	}
#endif

	/* The p3043-specific entries in the device structure. */
	dev->open = &p3043_open;
	dev->hard_start_xmit = &p3043_start_xmit;
	dev->stop = &p3043_close;
	dev->get_stats = &p3043_get_stats;
	dev->set_multicast_list = &set_rx_mode;

	return dev;
}



static int
p3043_open(struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	int ioaddr = dev->base_addr;
	int i;
        int j;    
//	int full_duplex;
        int FDXFlag,CRbak;
        int rx_ring_tmp,rx_ring_tmp1;
        int tx_ring_tmp,tx_ring_tmp1;
        int rx_bufs_tmp,rx_bufs_tmp1;
        int tx_bufs_tmp,tx_bufs_tmp1;

 
	/* Soft reset the chip. */
	//outb(CmdReset, ioaddr + ChipCmd);

#ifdef SA_SHIRQ
        //printk("hook int,[%d]",dev->irq); 
	if (request_irq(dev->irq, &p3043_interrupt, SA_SHIRQ,
					"RealTek p3043/39 Fast Ethernet", dev)) {
		return -EAGAIN;
	}
#else
	if (irq2dev_map[dev->irq] != NULL
		|| (irq2dev_map[dev->irq] = dev) == NULL
		|| dev->irq == 0
		|| request_irq(dev->irq, &p3043_interrupt, 0, "p3043")) {
		return -EAGAIN;
	}
#endif

	MOD_INC_USE_COUNT;

      tx_bufs_tmp = ( int )kmalloc(TX_RING_SIZE * PKT_BUF_SZ + 32, GFP_KERNEL);
      tx_ring_tmp = ( int )kmalloc(TX_RING_SIZE * sizeof(struct aln320_tx_desc) + 32, GFP_KERNEL);
      rx_bufs_tmp = ( int )kmalloc(RX_RING_SIZE * PKT_BUF_SZ + 32, GFP_KERNEL);
      rx_ring_tmp = ( int )kmalloc(RX_RING_SIZE * sizeof(struct aln320_rx_desc) +32, GFP_KERNEL);

      /* tune RD TD 32 byte alignment */              
        rx_ring_tmp1=(int)virt_to_bus((char *)rx_ring_tmp);
        j=(rx_ring_tmp1+32)&(~0x1f);       
      //  printk("txring[%d]",j);
        tp->rx_ring=(struct aln320_rx_desc *)bus_to_virt(j);
        
        tx_ring_tmp1=( int )virt_to_bus((char *)tx_ring_tmp);
        j=(tx_ring_tmp1+32)&(~0x1f);       
        tp->tx_ring=(struct aln320_tx_desc *)bus_to_virt(j);
        //printk("rxring[%x]",j);
 

        tx_bufs_tmp1=( int )virt_to_bus((char *)tx_bufs_tmp);
        j=(int )(tx_bufs_tmp1+32)&(~0x1f);       
        tx_bufs_tmp=(int)bus_to_virt(j);
        //printk("txb[%x]",j);

        rx_bufs_tmp1=( int )virt_to_bus((char *)rx_bufs_tmp);
        j=(int )(rx_bufs_tmp1+32)&(~0x1f);       
        rx_bufs_tmp=(int)bus_to_virt(j);
        //printk("rxb[%x][%x]",rx_bufs_tmp1,j);

        for (i=0 ;i<RX_RING_SIZE;i++){
              tp->rx_buffs[i]=(char *)rx_bufs_tmp; 
              //printk("r[%x]",tp->rx_buffs[i]);  
              rx_bufs_tmp+=1536;
        }  

        for (i=0 ;i<TX_RING_SIZE;i++){
              tp->tx_buffs[i]=(char *)tx_bufs_tmp; 
              //printk("t[%x]",tp->tx_buffs[i]);  
              tx_bufs_tmp+=1536;
        }  

        //printk("init ring");
	p3043_init_ring(dev);
      /*write TD RD Descriptor to MAC*/
        outl(virt_to_bus(tp->rx_ring),dwCurrentRxDescAddr);
	outl(virt_to_bus(tp->tx_ring),dwCurrentTxDescAddr);



       /* close IMR */
         outw(0x0000,byIMR0); 
    
       /* software reset */
         outb(CR1_SFRST,byCR1);
         MIIDelay();   
       /* set TCR RCR threshold */
         outb(0x06,byBCR0);
         outb(0x00,byBCR1);
         outb(0x2c,byRCR); 
         outb(0x60,byTCR); 
       /* Set Fulldupex */
         FDXFlag=QueryAuto(ioaddr);
         if(FDXFlag==1){
            outb(CFGD_CFDX,byCFGD);
            outw(CR_FDX,byCR0);
         }

       /*KICK NIC to WORK */
         CRbak=inw(byCR0);
         CRbak=CRbak&0xFFFB;/*not CR_STOP */
         outw((CRbak|CR_STRT|CR_TXON|CR_RXON|CR_DPOLL),byCR0);   


       /*set IMR to work */
         outw(IMRShadow,byIMR0);

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
	/* Set the timer to switch to check for link beat and perhaps switch
	   to an alternate media type. */
	init_timer(&tp->timer);
	tp->timer.expires = RUN_AT((24*HZ)/10);			/* 2.4 sec. */
	tp->timer.data = (unsigned long)dev;
	tp->timer.function = &p3043_timer;				/* timer handler */
	add_timer(&tp->timer);

	return 0;
}

static void p3043_timer(unsigned long data)
{
	struct device *dev = (struct device *)data;
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	//int ioaddr = dev->base_addr;
	int next_tick = 0;

	if (next_tick) {
		tp->timer.expires = RUN_AT(next_tick);
		add_timer(&tp->timer);
	}
}

static void p3043_tx_timeout(struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	//int ioaddr = dev->base_addr;

	dev->trans_start = jiffies;
	tp->stats.tx_errors++;
	return;
}


/* Initialize the Rx and Tx rings, along with various 'dev' bits. */

static int
p3043_start_xmit(struct sk_buff *skb, struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	int ioaddr = dev->base_addr;
	int entry;
        unsigned char CR1bak;
#ifndef final_version
	if (skb == NULL || skb->len <= 0) {
		printk("%s: Obsolete driver layer request made: skbuff==NULL.\n",
			   dev->name);
		dev_tint(dev);
		return 0;
	}
#endif
        //outw(0x0000,byIMR0);
        //printk("xmit");     

	/* Block a timer-based transmit from overlapping.  This could better be
	   done with atomic_swap(1, dev->tbusy), but set_bit() works as well. */

	if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
		if (jiffies - dev->trans_start < TX_TIMEOUT)
			return 1;
		p3043_tx_timeout(dev);
		return 1;
	}

         
	/* Calculate the next Tx descriptor entry. */
	entry = tp->cur_tx % TX_RING_SIZE;
        //whil1Ye(tp->tx_ring[entry].tx_status.bits.own_bit==1)
         //                   ;   
/*
        if(tp->tx_ring[entry].tx_status.bits.own_bit==1){
          for (i=0;i<1000000;i++) 
          {  
          if(tp->tx_ring[entry].tx_status.bits.own_bit==0)
                      break; 
          }       
        }                              
*/
	tp->tx_skbuff[entry] = skb;
        if(skb->len<60)
              skb->len=60;
        //printk("tx[%d],len=[%d]",entry,skb->len);
	//if ((long)skb->data & 3) {			/* Must use alignment buffer. */
		memcpy(tp->tx_buffs[entry], skb->data, skb->len);
		//outl(virt_to_bus(tp->tx_buf[entry]), ioaddr + TxAddr0 + entry*4);
	//} else
	//	outl(virt_to_bus(skb->data), ioaddr + TxAddr0 + entry*4);
	/* Note: the chip doesn't have auto-pad! */

	 tp->tx_ring[entry].tx_ctrl.bits.tx_buf_size=skb->len; 

         tp->tx_ring[entry].tx_status.bits.own_bit=1; 

             
         CR1bak=inb(byCR1);

         CR1bak=CR1bak|CR1_TDMD1;
         //printk("tdsw=[%x]",tp->tx_ring[entry].tx_status.lw);
         //printk("tdcw=[%x]",tp->tx_ring[entry].tx_ctrl.lw);
         //printk("tdbuf1=[%x]",tp->tx_ring[entry].buf_addr_1);
         //printk("tdbuf2=[%x]",tp->tx_ring[entry].buf_addr_2);
         //printk("td1=[%x]",inl(dwCurrentTDSE0));       
         //printk("td2=[%x]",inl(dwCurrentTDSE1));       
         //printk("td3=[%x]",inl(dwCurrentTDSE2));       
         //printk("td4=[%x]",inl(dwCurrentTDSE3));       

        outb(CR1bak,byCR1); 
        // tp->cur_tx ++;

        if(++tp->cur_tx - tp->dirty_tx < TX_RING_SIZE ){
          dev->tbusy=0;
        }else{
          tp->tx_full = 1;      	
        }      
	dev->trans_start = jiffies;
        //outw(IMRShadow,byIMR0);
        //dev_kfree_skb(tp->tx_skbuff[entry], FREE_WRITE);
        //tp->tx_skbuff[entry] = 0;
	return 0;
}

/* The interrupt handler does all of the Rx thread work and cleans up
   after the Tx thread. */
static void p3043_interrupt IRQ(int irq, void *dev_instance, struct pt_regs *regs)
{
#ifdef SA_SHIRQ
	struct device *dev = (struct device *)dev_instance;
#else
	struct device *dev = (struct device *)(irq2dev_map[irq]);
#endif
	struct p3043_private *tp;
	int ioaddr;
	int status;
        //int RCRbak,TCRbak,IMRbak;
        int CRbak;
        char CR0bak,CR1bak,FDXFlag,LNKFlag0;
        int txstatus;

	if (dev == NULL) {
		printk ("p3043_interrupt(): irq %d for unknown device.\n", irq);
		return;
	}

	ioaddr = dev->base_addr;
	tp = (struct p3043_private *)dev->priv;

	if (dev->interrupt)
		printk("%s: Re-entering the interrupt handler.\n", dev->name);

	dev->interrupt = 1;

                outw(0x0000,byIMR0);
                tp->cur_rx=tp->cur_rx % RX_RING_SIZE; 
                status=inw(byISR0);
		/* Acknowledge all of the current interrupt sources ASAP. */
                //printk("isr=%x",status);
                outw(status,byISR0);

                 if (status & (ISR_PRX|ISR_RXE|ISR_RU)) {
                    while(tp->rx_ring[tp->cur_rx].rx_status.bits.own_bit == 0 ){
                       p3043_rx(dev);
                       tp->rx_ring[tp->cur_rx].rx_status.bits.own_bit=1;
	               tp->cur_rx++;
                       tp->cur_rx=tp->cur_rx % RX_RING_SIZE; 
                    }
                 }

	    
                 if (status & (ISR_PTX|ISR_TXE|ISR_TU)) {  /* transmit packet */
			int dirty_tx;
	          for (dirty_tx = tp->dirty_tx; dirty_tx < tp->cur_tx; dirty_tx++) {
			int entry = dirty_tx % TX_RING_SIZE;

                        if(tp->tx_ring[entry].tx_status.bits.own_bit==1)
                                        break;  

                         txstatus=tp->tx_ring[entry].tx_status.lw; 

                          //printk("txstatus=%x",txstatus);
                        if((txstatus) & (TSR_ABNORMAL))
                        {
                               // printk("txstatus=%x",txstatus);
                         
                        if(txstatus & TSR_TERR){
                                //  printk("TERR");  
                        }  
                       
                          if(txstatus & (TSR_ABT|TSR_UDF)){
                                 // adjust TD pointer and TDMD1 and TXON  
                              
                               CR0bak=inb(byCR0);
                               CR0bak=CR0bak|CR0_TXON;
                               outb(CR0bak,byCR0);
                               CR1bak=inb(byCR1);
                               CR1bak=CR1bak|CR1_TDMD1;
                               outb(CR1bak,byCR1);
                               //printk("ABT");
                         }  

                         if(txstatus & TSR_OWC){
                         //     printk("OWC"); 
                        }  
                         if(txstatus & TSR_CRS){
                        //     printk("CRS"); 
                        }  
                       if(txstatus & TSR_CDH){
                       //     printk("CDH"); 
                       }  
                       if(txstatus & TSR_COLS){
                       //     printk("COLS");  
                       }  
          
                       //   txerrint(dev, txstatus);
                       } else
                          tp->tx_ring[entry].tx_status.lw=0x0000; 

			/* Free the original skb. */
			dev_kfree_skb(tp->tx_skbuff[entry], FREE_WRITE);
			tp->tx_skbuff[entry] = 0;

  	          }
#ifndef final_version
			if (tp->cur_tx - dirty_tx > TX_RING_SIZE) {
				printk("%s: Out-of-sync dirty pointer, %d vs. %d, full=%d.\n",
					   dev->name, dirty_tx, tp->cur_tx, tp->tx_full);
				dirty_tx += TX_RING_SIZE;
			}
#endif
			if (tp->tx_full && dev->tbusy
				&& dirty_tx > tp->cur_tx - TX_RING_SIZE) {
				/* The ring is no longer full, clear tbusy. */
				tp->tx_full = 0;
				dev->tbusy = 0;
				mark_bh(NET_BH);
			}

			tp->dirty_tx = dirty_tx;
                        //printk("cur=[%d],dir[%d]",tp->cur_tx,tp->dirty_tx);   
                }
                 

        /* link change process */    
        if (status & ISR_SRCI){
           if(inb(byMIISR) & MIISR_LNKFL) {
            //printk("link fail");
              WriteMII(0,9,1,ioaddr); 
              LNKFlag0=1; 
            }else{
              LNKFlag0=0;
              FDXFlag=QueryAuto(ioaddr);
              if (FDXFlag==1){
                 outb(inb(byCR1)|CR1_FDX,byCR1); 
              }else{
                 outb(inb(byCR1)&0xFB,byCR1); 
              }
             // printk("link change fdxflag=%d",FDXFlag); 
            }       
        } 
        



          

        if ((status) & (ISR_ABNORMAL)) {  /*abnormal isr */


           if (status& ISR_ABTI){
           // adjust TD pointer and TDMD1 and TXON  
              CR0bak=inb(byCR0);
              CR0bak=CR0bak|CR0_TXON|CR0_RXON|CR0_STRT;
              outb(CR0bak,byCR0);
              CR1bak=inb(byCR1);
              CR1bak=CR1bak|CR1_TDMD1;
              outb(CR1bak,byCR1);
    
           /* software reset */
              outb(CR1_SFRST,byCR1);
              MIIDelay();   

           /* set TCR RCR threshold */
              outb(0x06,byBCR0);
              outb(0x00,byBCR1);
              outb(0x2c,byRCR); 
              outb(0x60,byTCR); 

           /* Set Fulldupex */
              FDXFlag=QueryAuto(ioaddr);
              if(FDXFlag==1){
              outb(byCFGD,CFGD_CFDX);
              outw(CR_FDX,byCR0);
             }
             
           /*KICK NIC to WORK */
             CRbak=inw(byCR0);
             CRbak=CRbak&0xFFFB;/*not CR_STOP */
             outw((CRbak|CR_STRT|CR_TXON|CR_RXON|CR_DPOLL),byCR0);   
	                   
           } 



            if (status & ISR_NORBF){
               CR0bak=inb(byCR0);
               CR0bak=CR0bak|CR0_STRT|CR0_TXON|CR0_RXON;
               outb(CR0bak,byCR0);
            }
               if (status & ISR_RU){
               }
               if (status & ISR_TU){
               }
               if (status & ISR_CNT){
               }  
               if (status & ISR_TXE){
                  CRbak=inw(byCR0);
                  CRbak=CRbak&0xFFFB;
                  outw((CRbak|CR_TXON|CR_RXON|CR_STRT),byCR0);   
               }
               if (status & ISR_PKTRA){ /*software reset*/
               
               }
               if (status & ISR_BE){ /* kick NIC to work */
                  CRbak=inw(byCR0);
                  CRbak=CRbak&0xFFFB;
                  outw((CRbak|CR_TXON|CR_RXON|CR_DPOLL),byCR0);   
               }
           
          }

          dev->interrupt = 0;
          outw(IMRShadow,byIMR0);
          return;

}


static int
p3043_rx(struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	//int ioaddr = dev->base_addr;
        int rxstatus;        
        u16 rxsize; 
        unsigned char *skbtmp;    
        struct sk_buff *skb;
        rxstatus=tp->rx_ring[tp->cur_rx].rx_status.lw; 

        if((rxstatus&0x0300)!=0x0300)
                return 0;
        if(rxstatus & (RSR_ABNORMAL)){
                return 0;
                printk("rxerr[%x]",rxstatus);
        }

        rxsize=tp->rx_ring[tp->cur_rx].rx_status.bits.frame_length; 

        /* Malloc up new buffer, compatible with net-2e. */
	/* Omit the four octet CRC from the length. */

	skb = DEV_ALLOC_SKB(rxsize + 2);

	if (skb == NULL) {
		printk("%s: Memory squeeze, deferring packet.\n", dev->name);
		/* We should check that some rx space is free.
		   If not, free one and mark stats->rx_dropped++. */
		tp->stats.rx_dropped++;
	                return 0;
	}
	skb->dev = dev;
	skb_reserve(skb, 2);	/* 16 byte align the IP fields. */
        skbtmp=skb_put(skb,rxsize);    
	memcpy(skbtmp, tp->rx_buffs[tp->cur_rx], rxsize);

#if LINUX_VERSION_CODE >= 0x10300
	skb->protocol = eth_type_trans(skb, dev);
#else
	skb->len = rxsize;
#endif
	netif_rx(skb);
	tp->stats.rx_packets++;
        //printk("upsend");        

	return 0;

}

static int
p3043_close(struct device *dev)
{
	int ioaddr = dev->base_addr;
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	int i;

	dev->start = 0;
	dev->tbusy = 1;


	/* Disable interrupts by clearing the interrupt mask. */
	outw(0x0000, byIMR0);

	/* Stop the chip's Tx and Rx DMA processes. */
	outw(CR_STOP, byCR0);


	del_timer(&tp->timer);

#ifdef SA_SHIRQ
	free_irq(dev->irq, dev);
#else
	free_irq(dev->irq);
	irq2dev_map[dev->irq] = 0;
#endif


	for (i = 0; i < TX_RING_SIZE; i++) {
		if (tp->tx_skbuff[i])
			dev_kfree_skb(tp->tx_skbuff[i], FREE_WRITE);
		tp->tx_skbuff[i] = 0;
	        kfree(tp->tx_buffs[i]);
	}

	for (i = 0; i < RX_RING_SIZE; i++) {
	        kfree(tp->rx_buffs[i]);
	}
	kfree(tp->rx_ring);
	kfree(tp->tx_ring);

	MOD_DEC_USE_COUNT;

	return 0;
}

static struct enet_statistics *
p3043_get_stats(struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	//int ioaddr = dev->base_addr;

	/*if (dev->start) {
		tp->stats.rx_missed_errors += inl(ioaddr + RxMissed);
		outl(0, ioaddr + RxMissed);
	}*/

	return &tp->stats;
}

/* Set or clear the multicast filter for this adaptor.
   Note that we only use exclusion around actually queueing the
   new frame, not around filling tp->setup_frame.  This is non-deterministic
   when re-entered but still correct. */

/* The little-endian AUTODIN II ethernet CRC calculation.
   N.B. Do not use for bulk data, use a table-based routine instead.
   This is common code and should be moved to net/core/crc.c */
static unsigned const ethernet_polynomial_le = 0xedb88320U;
static inline unsigned ether_crc_le(int length, unsigned char *data)
{
	unsigned int crc = 0xffffffff;	/* Initial value. */
	while(--length >= 0) {
		unsigned char current_octet = *data++;
		int bit;
		for (bit = 8; --bit >= 0; current_octet >>= 1) {
			if ((crc ^ current_octet) & 1) {
				crc >>= 1;
				crc ^= ethernet_polynomial_le;
			} else
				crc >>= 1;
		}
	}
	return crc;
}


static void
#ifdef NEW_MULTICAST
set_rx_mode(struct device *dev)
#else
static void set_rx_mode(struct device *dev, int num_addrs, void *addrs);
#endif
{
	int ioaddr = dev->base_addr;
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	unsigned char mc_filter[8];		 /* Multicast hash filter */
	int i;

	if (dev->flags & IFF_PROMISC) {			/* Set promiscuous. */
		/* Unconditionally log net taps. */
		printk("%s: Promiscuous mode enabled.\n", dev->name);
		memset(mc_filter, 0xff, sizeof(mc_filter));
		outb(0x2e, byRCR);
	} else if ((dev->mc_count > 1000)  ||  (dev->flags & IFF_ALLMULTI)) {
		/* Too many to filter perfectly -- accept all multicasts. */
		memset(mc_filter, 0xff, sizeof(mc_filter));
		outb(0x2c, byRCR);
	} else if (dev->mc_count == 0) {
		outb(0x2c, byRCR);
		return;
	} else {
		struct dev_mc_list *mclist;

		memset(mc_filter, 0, sizeof(mc_filter));
		for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next)
			set_bit(ether_crc_le(ETH_ALEN, mclist->dmi_addr) & 0x3f,
					mc_filter);
	}
	/* ToDo: perhaps we need to stop the Tx and Rx process here? */
	if (memcmp(mc_filter, tp->mc_filter, sizeof(mc_filter))) {
		for (i = 0; i < 2; i++)
			outl(((u32 *)mc_filter)[i], byMAR0 + i*4);
		memcpy(tp->mc_filter, mc_filter, sizeof(mc_filter));
	}

	return;
}

#ifdef MODULE

/* An additional parameter that may be passed in... */
static int debug = -1;

int
init_module(void)
{
	int cards_found;

	if (debug >= 0)
		p3043_debug = debug;

	root_p3043_dev = NULL;
	cards_found = aln320_probe(0);
        //printk("card=%d",cards_found);
	return cards_found ? 0 : -ENODEV;
}

void
cleanup_module(void)
{
	struct device *next_dev;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_p3043_dev) {
		next_dev = ((struct p3043_private *)root_p3043_dev->priv)->next_module;
		unregister_netdev(root_p3043_dev);
		release_region(root_p3043_dev->base_addr, p3043_TOTAL_SIZE);
		kfree(root_p3043_dev);
		root_p3043_dev = next_dev;
	}
}


int QueryAuto(int ioaddr)
 {
     int   byMIIIndex;
     int   MIIReturn;
     int   FDXFlag;
     byMIIIndex=0x06;
     MIIReturn=ReadMII(byMIIIndex,ioaddr);
     if ((MIIReturn&0x01)==0){
     FDXFlag=0;
     return FDXFlag;    }

     byMIIIndex=0x05;
     MIIReturn=ReadMII(byMIIIndex,ioaddr);
     if ((MIIReturn&0x0140)==0){
     FDXFlag=0;
     return FDXFlag;    }

     byMIIIndex=0x04;
     MIIReturn=ReadMII(byMIIIndex,ioaddr);
     if ((MIIReturn&0x0140)==0){
     FDXFlag=0;
     return FDXFlag;    }

     FDXFlag=1;
     return FDXFlag;

 }


int ReadMII(int byMIIIndex,int ioaddr)
{
   int  ReturnMII;
   char byMIIAdrbak;
   char byMIICRbak;
   char byMIItemp;

   byMIIAdrbak=inb(byMIIAD);
   byMIICRbak=inb(byMIICR);
   outb(byMIICRbak&0x7f,byMIICR);
   MIIDelay();

   outb(byMIIIndex,byMIIAD);
   MIIDelay();

   outb(inb(byMIICR)|0x40,byMIICR);

   byMIItemp=inb(byMIICR);
   byMIItemp=byMIItemp&0x40;

  while(byMIItemp!=0)
   {
     byMIItemp=inb(byMIICR);
     byMIItemp=byMIItemp&0x40;
   }
   MIIDelay();

   ReturnMII=inw(wMIIDATA);

   outb(byMIIAdrbak,byMIIAD);
   outb(byMIICRbak,byMIICR);
   MIIDelay();

   return(ReturnMII);

}
void WriteMII(char byMIISetByte,char byMIISetBit,char byMIIOP,int ioaddr)
{
   int  ReadMIItmp;
   int  MIIMask;
   char byMIIAdrbak;
   char byMIICRbak;
   char byMIItemp;

   
   byMIIAdrbak=inb(byMIIAD);

   byMIICRbak=inb(byMIICR);
   outb(byMIICRbak&0x7f,byMIICR);
   MIIDelay();
   outb(byMIISetByte,byMIIAD);
   MIIDelay();

   outb(inb(byMIICR)|0x40,byMIICR);

   byMIItemp=inb(byMIICR);
   byMIItemp=byMIItemp&0x40;

   while(byMIItemp!=0)
   {
     byMIItemp=inb(byMIICR);
     byMIItemp=byMIItemp&0x40;
   }
   MIIDelay();

   ReadMIItmp=inw(wMIIDATA);
   MIIMask=0x0001;
   MIIMask=MIIMask<<byMIISetBit;


   if (byMIIOP==0){
       MIIMask=~MIIMask;
       ReadMIItmp=ReadMIItmp&MIIMask;
   }else{
       ReadMIItmp=ReadMIItmp|MIIMask;

   }
   outw(ReadMIItmp,wMIIDATA);
   MIIDelay();

   outb(inb(byMIICR)|0x20,byMIICR);
   byMIItemp=inb(byMIICR);
   byMIItemp=byMIItemp&0x20;

   while(byMIItemp!=0)
   {
     byMIItemp=inb(byMIICR);
     byMIItemp=byMIItemp&0x20;
   }
   MIIDelay();

   outb(byMIIAdrbak&0x7f,byMIIAD);
   outb(byMIICRbak,byMIICR);
   MIIDelay();

}

void MIIDelay(void)
{
 int i;
 for(i=0;i<0x7fff;i++){
    inb(0x61);
    inb(0x61);
    inb(0x61);
    inb(0x61);
 }
}


void post_reset_delay(void)
{

    uint initial_time, current_time;

    /* need to wait 10 mini seconds */

    initial_time = read_tick_counter();

    /* 8254 time tick gives 419 nsec per count. We have to delay a delta
       count of 10,000,000ns/419ns = 23866. We will round it up to 23866
     */

    while (1)
        {
        current_time = read_tick_counter();
        if ((initial_time - current_time) >= 23866)   /* 10 mini second */
            {
            return;
            }
        }
}



uint read_tick_counter(void)
{
    uint current_time = 0;

    /* Program 8254 to latch T0's count */
    outb(0x43, 0x6);
    /* Read the counter */
    current_time = inb(0x40);
    current_time = current_time | (inb(0x40) << 8);
    return current_time;

}

/* Initialize the Rx and Tx rings, along with various 'dev' bits. */
static void
p3043_init_ring(struct device *dev)
{
	struct p3043_private *tp = (struct p3043_private *)dev->priv;
	int i;

	tp->tx_full = 0;
	tp->cur_rx = tp->cur_tx = 0;
	tp->dirty_rx = tp->dirty_tx = 0;

	for (i = 0; i < RX_RING_SIZE; i++) {
	                
                tp->rx_ring[i].rx_status.bits.own_bit = 1;
                tp->rx_ring[i].rx_ctrl.bits.rx_buf_size =1536;
  
		tp->rx_ring[i].buf_addr_1 = virt_to_bus(tp->rx_buffs[i]);
		tp->rx_ring[i].buf_addr_2 = virt_to_bus(&tp->rx_ring[i+1]);
                //printk("[%d]buf1=%x,buf2=%x",i,tp->rx_ring[i].buf_addr_1,tp->rx_ring[i].buf_addr_2);
	}
	/* Mark the last entry as wrapping the ring. */ 
//	tp->rx_ring[i-1].rx_ctrl.bits.rx_buf_size =1518;
	tp->rx_ring[i-1].buf_addr_2 = virt_to_bus(&tp->rx_ring[0]);
        //printk("[%d]buf1=%x,buf2=%x",i-1,tp->rx_ring[i-1].buf_addr_1,tp->rx_ring[i-1].buf_addr_2);

	/* The Tx buffer descriptor is filled in as needed, but we
	   do need to clear the ownership bit. */

	for (i = 0; i < TX_RING_SIZE; i++) {

                tp->tx_ring[i].tx_status.lw = 0;
                tp->tx_ring[i].tx_ctrl.lw =0x00e08000;
                tp->tx_ring[i].buf_addr_1 = virt_to_bus(tp->tx_buffs[i]);
		tp->tx_ring[i].buf_addr_2 = virt_to_bus(&tp->tx_ring[i+1]);
//        printk("[%d]buf1=%x,buf2=%x",i,tp->tx_ring[i].buf_addr_1,tp->tx_ring[i].buf_addr_2);
	}

                 tp->tx_ring[i-1].buf_addr_2= virt_to_bus(&tp->tx_ring[0]);

 //       printk("[%d]buf1=%x,buf2=%x",i,tp->tx_ring[i-1].buf_addr_1,tp->tx_ring[i-1].buf_addr_2);
}




#endif  /* MODULE */
/*
 * Local variables:
 *  compile-command: "gcc -DMODVERSIONS  -DMODULE -D__KERNEL__ -Wall -Wstrict-prototypes -O6 -c aln320.c"
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
